#include "dynarray.h"


/**************************************
 *  BASE DYNAMIC ARRAY IMPLEMENTATION
 *************************************/

const int DEFAULT_DYNARRAY_CAPACITY = 2;

void dynarray_init(dynarray_t *a)
{
    /* allocate space for data */
    a->array = (data_t*)malloc(DEFAULT_DYNARRAY_CAPACITY * sizeof(data_t));
    if (a->array == NULL) {
        fprintf(stderr, "ERROR: Cannot allocate memory for dynamic array.");
        exit(-1);
    }

    /* initialize other metadata */
    a->capacity = DEFAULT_DYNARRAY_CAPACITY;
    a->size = 0;
    a->front = 0;
}

void dynarray_expand(dynarray_t *a, size_t new_capacity)
{
    /* allocate new array */
    data_t *new_array = (data_t*)malloc(new_capacity * sizeof(data_t));
    if (!new_array) {
        fprintf(stderr, "ERROR: Cannot allocate memory for dynamic array.");
        exit(-1);
    }

    /* copy items to new array */
    for (size_t i = 0; i < a->size; i++) {
        new_array[i] = a->array[(a->front + i) % a->capacity];
    }

    /* delete old array and clean up */
    free(a->array);
    a->array = new_array;
    a->capacity = new_capacity;
    a->front = 0;
}

void dynarray_insert(dynarray_t *a, size_t idx, data_t item)
{
    if (a->size+1 > a->capacity) {
        dynarray_expand(a, a->capacity*2);          /* expand if necessary */
    }
    for (size_t i = idx+1; i < a->size+1; i++) {    /* shift items right */
        a->array[i] = a->array[i-1];
    }
    a->array[idx] = item;                           /* insert new item */
    a->size++;                                      /* update size metadata */
}

data_t dynarray_remove(dynarray_t *a, size_t idx)
{
    assert(idx < a->size);
    data_t temp = a->array[idx];                /* save item to be deleted */
    for (size_t i=idx; i<a->size-1; i++) {      /* shift items left */
        a->array[i] = a->array[i+1];
    }
    a->size--;                                  /* update size metadata */
    return temp;                                /* return deleted item */
}

void dynarray_set(dynarray_t *a, size_t idx, data_t item)
{
    assert(idx < a->size);
    a->array[idx] = item;
}

data_t dynarray_get(dynarray_t *a, size_t idx)
{
    assert(idx < a->size);
    return a->array[idx];
}

size_t dynarray_size(dynarray_t *a)
{
    return a->size;
}

bool dynarray_is_empty(dynarray_t *a)
{
    return (a->size == 0);
}

void dynarray_free(dynarray_t *a)
{
    free(a->array);
    a->array    = NULL;
    a->capacity = 0;
    a->size     = 0;
}

/**************************************
 *          SEARCH
 **************************************/

 bool dynarray_contains(dynarray_t *a, data_t item)
 {
     for (size_t i = 0; i < a->size; i++) {
         if (a->array[(a->front + i) % a->capacity] == item) {
             return true;
         }
     }
     return false;
 }


/**************************************
 *           STACK INTERFACE
 *************************************/

void dynarray_push(dynarray_t *a, data_t item)
{
    if (a->size+1 > a->capacity) {
        dynarray_expand(a, a->capacity*2);          /* expand if necessary */
    }
    a->array[a->size++] = item;
}

data_t dynarray_pop(dynarray_t *a)
{
    assert(a->size > 0);
    return a->array[--(a->size)];
}

data_t dynarray_top(dynarray_t *a)
{
    assert(a->size > 0);
    return a->array[a->size-1];
}


/**************************************
 *           QUEUE INTERFACE
 *************************************/

void dynarray_enqueue(dynarray_t *a, data_t item)
{
    if (a->size+1 > a->capacity) {
        dynarray_expand(a, a->capacity*2);          /* expand if necessary */
    }
    a->array[(a->front + a->size) % a->capacity] = item;
    a->size++;
}

data_t dynarray_dequeue(dynarray_t *a)
{
    assert(a->size > 0);
    data_t value = a->array[a->front];
    a->front = (a->front+1) % a->capacity;
    a->size--;
    return value;
}

data_t dynarray_front(dynarray_t *a)
{
    assert(a->size > 0);
    return a->array[a->front];
}
